#include "../../global.h"

#include "../../usb.h"
#include "../../usb_api.h"

#include <mint/osbind.h> // Super + SuperToUser

//#define TOSONLY

// ktypes.h defines u_int[size]_t but not uint[size]_t
typedef u_int8_t uint8_t;
typedef u_int16_t uint16_t;
typedef u_int32_t uint32_t;

#include "mouse_report_desc_parser.h"

#ifdef TOSONLY
#include "MWheel.h"
#define MSG_VERSION "TOS DRIVERS"
#else
#define MSG_VERSION "FreeMiNT DRIVERS"
#endif

#define MSG_BUILDDATE	__DATE__

#define MSG_BOOT	\
	"\033p USB mouse class driver " MSG_VERSION " \033q\r\n"

#define MSG_GREET	\
	"Ported, mixed and shaken by Alan Hourihane.\r\n" \
	"HID Report DescriptorParser\r\n"\
	"    & MWheel-Extension by Armin Diedering. \r\n" \
	"Compiled " MSG_BUILDDATE ".\r\n\r\n"


/****************************************************************************/
/*
 * BEGIN kernel interface
 */

#ifndef TOSONLY
struct kentry *kentry;

void mouse_poll_thread (void *);
void mouse_poll (PROC * proc, long dummy);
#else
extern unsigned long _PgmSize;

/*
 * old handler
 */
extern void (*old_ikbd_int) (void);

/*
 * interrupt wrapper routine
 */
extern void interrupt_ikbd (void);
#endif

struct kbdvbase
{
	long midivec;
	long vkbderr;
	long vmiderr;
	long statvec;
	long mousevec;
	long clockvec;
	long joyvec;
	long midisys;
	long ikbdsys;
	short drvstat;			/* Non-zero if a packet is currently
							 * transmitted. */
};
typedef struct kbdvbase KBDVEC;

KBDVEC *vector;
static char mouse_packet[6];
void _cdecl send_packet (long func, char *buf, char *bufend);
void _cdecl fake_hwint(void);

#ifndef TOSONLY
void _cdecl send_wheel(short wheel, short wclicks);
extern unsigned long vdi_buttons;
void vdi_btnv(void);
extern void(*old_vdi_btnv)(void);
extern void(*vdi_wheelv)(void);
extern void trap2(void);
extern void(*old_trap2)(void);
#endif
struct usb_module_api *api;

#ifdef TOSONLY

MWHL_Cookie *mwhl = 0;
short mwhl_value[16];

#endif

void mouse_int (void);

/*
 * END kernel interface
 */
/****************************************************************************/

/*
 * USB device interface
 */

static long mouse_ioctl (struct uddif *, short, long);
static long mouse_disconnect (struct usb_device *dev);
static long mouse_probe (struct usb_device *dev, unsigned int ifnum);

static char lname[] = "USB mouse class driver\0";

static struct uddif mouse_uif = {
	0,                          /* *next */
	USB_API_VERSION,            /* API */
	USB_DEVICE,                 /* class */
	lname,                      /* lname */
	"mouse",                    /* name */
	0,                          /* unit */
	0,                          /* flags */
	mouse_probe,                /* probe */
	mouse_disconnect,           /* disconnect */
	0,                          /* resrvd1 */
	mouse_ioctl,                /* ioctl */
	0,                          /* resrvd2 */
};


struct mse_data
{
	struct usb_device *pusb_dev;        /* this usb_device */
	unsigned char ep_in;        /* in endpoint */
	unsigned char ep_out;       /* out ....... */
	unsigned char ep_int;       /* interrupt . */
	long *irq_handle;            /* for USB int requests */
	unsigned long irqpipe;      /* pipe for release_irq */
	unsigned char irqmaxp;      /* max packed for irq Pipe */
	unsigned char irqinterval;  /* Intervall for IRQ Pipe */
	unsigned char report_len;
	char buttons;
	char data[64];
	long move_and_scroll[4]; // moveX, moveY, scrollWheelV, scrollWheelH
#define moveX move_and_scroll[0]
#define moveY move_and_scroll[1]
#define scrollWheelV move_and_scroll[2]
#define scrollWheelH move_and_scroll[3]
#define scrollWheel(idx) move_and_scroll[2+idx]

	HID_ENTRY hid_entries[9];
};

static const HID_ENTRY hid_boot_entries[9] = {
	{ 0, 0, 1, 0, 1 },			// button 1
	{ 0, 1, 1, 0, 1 },			// button 2
	{ 0, 2, 1, 0, 1 },			// button 3
	{ 0, 3, 1, 0, 1 },			// button 4
	{ 0, 4, 1, 0, 1 },			// button 5
	{ 0, 8, 8, -128, 127 },		// X
	{ 0, 16, 8, -128, 127 },	// Y
	{ 0, 0, 0, 0, 0 },			// wheel 1
	{ 0, 0, 0, 0, 0 },			// wheel 2
};


static struct mse_data mse_data;

/*
* --- Helper functions
* --------------------------------------------------
*/

static char clampMouseMove(long *remaining) {
	long ret;
	if (*remaining > 127)
		ret = 127;
	else if (*remaining < -128)
		ret = -128;
	else
		ret = (char)*remaining;
	*remaining -= ret;
	return ret;
}

#ifndef TOSONLY

typedef union {
	short *s;
	void(**f)(void);
} ptr_u;

static void vdi_vex(short op, void(*fnc)(void), void(**old_fnc)(void)) {
	short vdi_contrl[12] = { 0, 0, 0, 0, 0, 0, 0 /* vdi handle 0 */ , 0 , 0, 0, 0, 0 };
	short *vdi_pb[5] = { vdi_contrl,0,0,0,0 };
	vdi_contrl[0] = op;
	*((ptr_u)&vdi_contrl[7]).f = fnc;
	__asm__ volatile (
		"move.l %0, d1\n"
		"	moveq.l #115, d0\n"
		"	trap #2"
		: /* no output */: "g"(vdi_pb), "r"(op) : "d0", "d1", "d2", "a0", "a1", "a2", "memory"
		);
	*old_fnc = *((ptr_u)&vdi_contrl[9]).f;
}

#endif

/*
 * --- Inteface functions
 * --------------------------------------------------
 */

static long _cdecl
mouse_ioctl (struct uddif *u, short cmd, long arg)
{
	return E_OK;
}

/*
 * -------------------------------------------------------------------------
 */

/*******************************************************************************
 *
 *
 */
static long
mouse_disconnect (struct usb_device *dev)
{
	if (dev == mse_data.pusb_dev)
	{
		mse_data.pusb_dev = NULL;
#ifndef TOSONLY
		wake (WAIT_Q, (long) &mouse_poll_thread);
#endif
	}

	return 0;
}

static long
usb_mouse_irq (struct usb_device *dev)
{
	return 0;
}

void
mouse_int (void)
{
	long i, changed = 0;
	long actlen = 0;
	long r;
	char *p, buttons = 0;
//	unsigned short sr;
	static char button_bits[] = { 1 << 1, 1 << 0, 1 << 2, 1 << 3, 1 << 4 };

	if (mse_data.pusb_dev == NULL)
		return;

#if 0
	usb_submit_int_msg (mse_data.pusb_dev,
						mse_data.irqpipe,
						mse_data.data,
						mse_data.irqmaxp > 8 ? 8 : mse_data.irqmaxp,
						USB_CNTL_TIMEOUT * 5);
#else
	r = usb_bulk_msg (mse_data.pusb_dev,
					  mse_data.irqpipe,
					  mse_data.data,
					  mse_data.report_len > 64 ? 64 : mse_data.report_len,
					  &actlen, USB_CNTL_TIMEOUT * 5, 1);
#endif

	if ((r != 0) || (actlen < 3))
		return;
	for (i = 0; i < 5; ++i) {
		if (mse_data.hid_entries[i].size) {
			p = mse_data.data;
			if ((mse_data.hid_entries[i].reportID == 0) || (mse_data.hid_entries[i].reportID == *p++)) {
				u_int16_t currBit = mse_data.hid_entries[i].offset;
				if (p[currBit / 8] & (1 << (currBit % 8)))
					buttons |= button_bits[i];
			}
		}
	}
	changed = mse_data.buttons ^ buttons;
	mse_data.buttons = buttons;
	for (i = 5; i < 9; ++i) {
		if (mse_data.hid_entries[i].size) {
			p = mse_data.data;
			if ((mse_data.hid_entries[i].reportID == 0) || (mse_data.hid_entries[i].reportID == *p++)) {
				u_int16_t currBit = mse_data.hid_entries[i].offset;
				u_int16_t bitsRem = mse_data.hid_entries[i].size;
				u_int32_t bitMask = (1 << 0);
				int32_t value = 0;

				while (bitsRem--) {
					if (p[currBit / 8] & (1 << (currBit % 8)))
						value |= bitMask;
					currBit++;
					bitMask <<= 1;
				}
				if (mse_data.hid_entries[i].logicalMin < 0 || mse_data.hid_entries[i].logicalMax < 0) {
					// logicalMin or lagicalMax are negative -> signExtent value
					bitsRem = 32 - mse_data.hid_entries[i].size;
					if (bitsRem > 0) value = (value << bitsRem) >> bitsRem;
				}
				mse_data.move_and_scroll[i - 5] = value;
				if (value) changed |= 1 << (8 + i - 5);
			}
		}
	}

#ifdef TOSONLY
	if (changed & 0x0303) { // moveX or moveY or left button or right button
		mouse_packet[0] = 0xF8 | (mse_data.buttons & 3);
		do {
			mouse_packet[1] = clampMouseMove(&mse_data.moveX);
			mouse_packet[2] = clampMouseMove(&mse_data.moveY);
		} while (mse_data.moveX || mse_data.moveY);
	}
	if (mwhl) { // MWheel?
		if (changed & 0x001C) { // any of button 3-5
			mwhl->mwhl_service(MWHL_SET_BUTTON, 0x001C, mse_data.buttons & 0x1C);
		}
		if (changed & 0x0C00) { // Wheel 1 or 2
			for (i = 0; i < 2; ++i) {
				if (changed & (0x400 << i)) {
					mwhl_value[i] = mse_data.scrollWheel(i);
					mse_data.scrollWheel(i) = 0;
				}
			}
			mwhl->mwhl_wheeled(mwhl_value);
		}
	}
#else
	if (changed & 0x031F) { // moveX or moveY or any button
		vdi_buttons = mse_data.buttons & 0x1c;
		mouse_packet[0] = 0xF8 | (mse_data.buttons & 3);
		do {
			mouse_packet[1] = clampMouseMove(&mse_data.moveX);
			mouse_packet[2] = clampMouseMove(&mse_data.moveY);
			send_packet(vector->mousevec, mouse_packet, mouse_packet + 3);
		} while (mse_data.moveX || mse_data.moveY);
	}
	for (i = 0; i < 2; ++i) {
		if (changed & (0x400 << i)) {
			send_wheel(i, mse_data.scrollWheel(i));
			mse_data.scrollWheel(i) = 0;
		}
	}
#endif
	if (changed) fake_hwint();

}

#ifndef TOSONLY
void
mouse_poll (PROC * proc, long dummy)
{
	wake (WAIT_Q, (long) &mouse_poll_thread);
}

void
mouse_poll_thread (void *dummy)
{
	p_setpriority(0,0,-20);

	while (mse_data.pusb_dev)
	{
		mouse_int ();
		addroottimeout (20, mouse_poll, 0);
		sleep (WAIT_Q, (long) &mouse_poll_thread);
	}

	kthread_exit (0);
}
#endif


/*******************************************************************************
 *
 *
 */
__attribute__((used)) static long
mouse_probe (struct usb_device *dev, unsigned int ifnum)
{
	struct usb_interface *iface;
	struct usb_endpoint_descriptor *ep_desc;
	unsigned char *usb_hid_report_descriptor;
	unsigned char *tempBuff;
	long usb_hid_report_descriptor_size;
	int have_bootprotocol = 1;
	int usb_hid_report_descriptor_ok = 0;

	/*
	 * Only one mouse at time
	 */
	if (mse_data.pusb_dev)
	{
		return -1;
	}

	if (dev == NULL)
	{
		return -1;
	}

	usb_disable_asynch (1);     /* asynch transfer not allowed */

	/*
	 * let's examine the device now
	 */
	iface = &dev->config.if_desc[ifnum];
	if (!iface)
	{
		return -1;
	}

	if (iface->desc.bInterfaceClass != USB_CLASS_HID)
	{
		return -1;
	}

	if (iface->desc.bInterfaceSubClass != USB_SUB_HID_BOOT)
	{
		have_bootprotocol = 0; // can't fall back to boot protocol
		//return -1;
	}

	if (have_bootprotocol && iface->desc.bInterfaceProtocol != 2)
	{
		return -1; // have boot protocol but isn't a mouse
	}

	if (iface->desc.bNumEndpoints != 1)
	{
		return -1;
	}

	ep_desc = &iface->ep_desc[0];
	if (!ep_desc)
	{
		return -1;
	}

	if ((ep_desc->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK) == USB_ENDPOINT_XFER_INT)
	{
		mse_data.ep_int = ep_desc->bEndpointAddress & USB_ENDPOINT_NUMBER_MASK;
		mse_data.irqinterval = ep_desc->bInterval;
	}
	else
	{
		return -1;
	}

	// allocate memory for descriptor and temporary Buffer for parser
	usb_hid_report_descriptor = (unsigned char*)kmalloc(2048 + mouse_hid_report_descriptor_parser(0, 0, 0, 0));
	if (!usb_hid_report_descriptor) {
		return -1;
	}
	tempBuff = usb_hid_report_descriptor + 2048;

	// read the hid report descriptor
	usb_hid_report_descriptor_size = usb_get_class_descriptor(dev, iface->desc.bInterfaceNumber, 0x22, 0, usb_hid_report_descriptor, 2048);

	if (usb_hid_report_descriptor_size > 0) {
#ifdef USB_MOUSE_WRITE_REPORT_DESCRIPTOR
		int32_t handle = Fcreate("C:\\mouse.hid", 0);
		if (handle > 0) {
			Fwrite(handle, usb_hid_report_descriptor_size, usb_hid_report_descriptor);
			Fclose(handle);
		}
#endif
		int32_t err;
		if ((err = mouse_hid_report_descriptor_parser(usb_hid_report_descriptor, usb_hid_report_descriptor_size, mse_data.hid_entries, tempBuff)) == 0) {
			short i;
			usb_hid_report_descriptor_ok = 1;
			mse_data.report_len = 8;
			for (i = 0; i < 8; ++i) {
				if (mse_data.hid_entries[i].size) {
					u_int16_t len = (mse_data.hid_entries[i].offset + mse_data.hid_entries[i].size + 7) / 8;
					if (mse_data.hid_entries[i].reportID) ++len;
					if (mse_data.report_len < len)
						mse_data.report_len = len;
				}
			}
		} else {
			DEBUG(("%s:%i:%s:hid device %i iface %i hid-parser err: %i", __FILE__, __LINE__, __FUNCTION__, (int)dev->devnum, ifnum, (int)err));
			if (have_bootprotocol) {
				DEBUG(("%s:%i:%s:fall back to boot protocol", __FILE__, __LINE__, __FUNCTION__));
				memcpy(mse_data.hid_entries, hid_boot_entries, sizeof(hid_boot_entries));
				mse_data.report_len = 8;
			}
		}
	}

	kfree(usb_hid_report_descriptor);
	if (!(have_bootprotocol | usb_hid_report_descriptor_ok)) {
		return -1;
	}

	mse_data.pusb_dev = dev;

	mse_data.irqinterval = (mse_data.irqinterval > 0) ? mse_data.irqinterval : 255;
	mse_data.irqpipe = usb_rcvintpipe (mse_data.pusb_dev, (long) mse_data.ep_int);
	mse_data.irqmaxp = usb_maxpacket (dev, mse_data.irqpipe);
	dev->irq_handle = usb_mouse_irq;

	usb_set_protocol(dev, iface->desc.bInterfaceNumber, usb_hid_report_descriptor_ok);

	// if(mse_data.irqmaxp < 6)
	//usb_set_protocol(dev, iface->desc.bInterfaceNumber, 0); /* boot */
	// else
	//usb_set_protocol (dev, iface->desc.bInterfaceNumber, 1);    /* report */

	usb_set_idle (dev, iface->desc.bInterfaceNumber, 0, 0);     /* report
                                                                 * infinite
                                                                 */

#ifndef TOSONLY
	long r = kthread_create (get_curproc (), mouse_poll_thread, NULL, NULL,
							 "mousepoll");

	if (r)
	{
		return 0;
	}
#endif

	return 0;
}

#ifdef TOSONLY
static inline MWHL_Cookie *
get_mwhl_cookie (void)
{
	MWHL_Cookie *mwhl_api;
	long ret;
	struct cookie *cjar;

	ret = Super(0L);

	mwhl_api = NULL;
	cjar = *CJAR;

	while (cjar->tag)
	{
		if (cjar->tag == MWHL_COOKIE)
		{
			mwhl_api = (MWHL_Cookie *)cjar->value;

			SuperToUser(ret);

			return mwhl_api;
		}

		cjar++;
	}

	SuperToUser(ret);

	return NULL;
}
#endif


#ifdef TOSONLY
int init (void);
int
init (void)
#else
long _cdecl init (struct kentry *, struct usb_module_api *, long, long);
long _cdecl
init (struct kentry *k, struct usb_module_api *uapi, long arg, long reason)
#endif
{
	long ret;

#ifndef TOSONLY
	kentry = k;
	api = uapi;

	if (check_kentry_version ())
		return -1;
#endif

	c_conws (MSG_BOOT);
	c_conws (MSG_GREET);
	DEBUG (("%s: enter init", __FILE__));

#ifdef TOSONLY
	/*
	 * GET _USB COOKIE to REGISTER
	 */
	api = get_usb_cookie ();
	if (!api)
	{
		(void) Cconws ("MOUSE failed to get _USB cookie\r\n");
		return -1;
	}

	/*
	* GET MWHL COOKIE
	*/
	mwhl = get_mwhl_cookie();
	if (mwhl) {
		int i;
		mwhl->mwhl_wheels |= 1;
		for (i = 0; i < 16; ++i) mwhl_value[i] = 0;
		c_conws("MWheel found.\r\n");
	}

#endif

	ret = udd_register (&mouse_uif);

	if (ret)
	{
		DEBUG (("%s: udd register failed!", __FILE__));
		return 1;
	}

	DEBUG (("%s: udd register ok", __FILE__));

	vector = (KBDVEC *) Kbdvbase ();

#ifndef TOSONLY

	vdi_vex(125, vdi_btnv, &old_vdi_btnv); // hook vex_btnv
	old_trap2 = Setexc(0x88 / 4, trap2);  // hook vdi trap
//	vdi_vex(134, vdi_ex_wheelv, &vdi_ex_wheelv);  // get vdi_ex_wheelv
//	vdi_vex(134, vdi_ex_wheelv, (void(**)(void))&ret);  // get vdi_ex_wheelv

#endif

#ifdef TOSONLY
#if 0
	old_ikbd_int = Setexc (0x114/4, (long) interrupt_ikbd);
#else
	{
		ret = Super (0L);
		old_ikbd_int = (void *) *(volatile unsigned long *) 0x400;
		*(volatile unsigned long *) 0x400 = (unsigned long) interrupt_ikbd;
		SuperToUser (ret);
	}
#endif
	c_conws ("USB mouse driver installed.\r\n");

	Ptermres (_PgmSize, 0);
#endif

	return 0;
}
